import numpy as np

def qc(df, obs=0.3):
    '''Filt data by observed rate (obs_num / all_num)'''
    if obs=='none' or obs==0:
        obs=0.01
    df = df.replace(0, np.nan)
    count_s = df.count()/len(df.index)
    count_s = count_s[count_s>=obs]
    return df[count_s.index]


import miidl
from .methods import default, none

def buildmodel(X_train=None, y_train=None, X_test=None, y_test=None, method='default', width=None, C=None, epochs=5):
    return getattr(miidl.modeling, method)(X_train, y_train, X_test, y_test, width, C, epochs)

call = {
    'default': default, 
    # 'custom': custom, 
    'none': none
}


import miidl
from .methods import log2, loge, log2p, logep, zscore, median, mean, none

def normalize(df, method='log2p'):
    return getattr(miidl.normalization, method)(df)

call = {
    'log2': log2,
    'loge': loge,
    'log2p': log2p, 
    'logep': logep, 
    'zscore': zscore, 
    'median': median, 
    'mean': mean, 
    'none': none
}


from .IO import read as _read
from .QC import qc as _qc
from .normalization import normalize as _normalize
from .imputation import impute as _impute
from .reshaping import reshape as _reshape
from .modeling import buildmodel as _buildmodel
from .interpretation import explain as _explain
import pandas as pd
import numpy as np
import torch


try:
    from captum.attr import visualization as viz
except:
    import sys
    print("The module 'captum' is not found, please install first.")
    print("\tconda install captum -c pytorch")
    sys.exit(0)

def _is_df(a):
    return True if type(a) is pd.DataFrame else False

def dataset_split(dataset, labels, train_proportion=0.7):
    n_all = len(labels)
    n_select = round(n_all * train_proportion)
    idx_all = range(n_all)
    idx_train = np.random.choice(n_all, size=n_select, replace=False)
    idx_test = list(set(idx_all) - set(idx_train))
    return dataset[idx_train], labels[idx_train], dataset[idx_test], labels[idx_test], idx_test

class MData:
    def __init__(self):
        self.full_data = None
        self.train_data = None
        self.test_data = None
        self.full_X = None
        self.full_y = None
        self.train_X = None
        self.train_y = None
        self.test_X = None
        self.test_y = None
        self.model = None
        self.attributions = None
        self.features = None
        self.num2label = {}
        self.label2num = {}
        self.importances = None
        self.full_y_sample_id = None
        self.test_y_sample_id = None

    def __repr__(self) -> str:
        if _is_df(self.full_data):
            print(self.full_data)
            return str(self.full_data.shape)
        if _is_df(self.train_data):
            print(self.train_data)
            if _is_df(self.test_data):
                print(self.test_data)
                return 'Train: '+str(self.train_data.shape)+'; Test: '+str(self.test_data.shape)
        return '0'

    def read(self, fname, role='all', group_col='Group'):
        if role=='all':
            self.full_data = _read(fname)
            self.full_y = self.full_data[group_col]
            self.full_X = self.full_data.drop(columns=group_col)
            self.full_data = self.full_data
        elif role=='train':
            self.train_data = _read(fname)
            self.train_y = self.train_data[group_col]
            self.train_X = self.train_data.drop(columns=group_col)
            self.train_data = self.train_data
        elif role=='test':
            self.test_data = _read(fname)
            self.test_y = self.test_data[group_col]
            self.test_X = self.test_data.drop(columns=group_col)
            self.test_data = self.test_data
        else:
            print(f"Illegal role: {role}!")
    
    def qc(self, obs=0.3):
        bf, af = 0, 0
        if _is_df(self.full_data):
            bf = len(list(self.full_X))
            self.full_X = _qc(self.full_X, obs)
            af = len(list(self.full_X))
        if _is_df(self.train_data):
            if _is_df(self.test_data):
                combined_X = pd.concat([self.train_X, self.test_X], sort=False)
                bf = len(list(combined_X))
                combined_X = _qc(combined_X, obs)
                af = len(list(combined_X))
                self.train_X = combined_X.loc[self.train_y.index,:]
                self.test_X = combined_X.loc[self.test_y.index,:]
        print(f'Number of features: from {bf} to {af}.')
    
    def normalize(self, method='log2p'):
        if _is_df(self.full_data):
            self.full_X = _normalize(self.full_X, method)
        if _is_df(self.train_data):
            if _is_df(self.test_data):
                self.train_X = _normalize(self.train_X, method)
                self.test_X = _normalize(self.test_X, method)
    
    def impute(self, method='none'):
        if _is_df(self.full_data):
            self.full_X = _impute(self.full_X, method)
        if _is_df(self.train_data):
            if _is_df(self.test_data):
                self.train_X = _impute(self.train_X, method)
                self.test_X = _impute(self.test_X, method)
    
    def reshape(self, method='auto', train_proportion=0.7):
        if _is_df(self.full_data):
            full_X, self.features = _reshape(self.full_X, method)
            uniq_label = list(set(self.full_y))
            self.full_y_sample_id = self.full_y.index
            self.full_y = np.array([uniq_label.index(i) for i in self.full_y])
            for i, v in enumerate(uniq_label):
                self.num2label[i] = v
                self.label2num[v] = i
            self.train_X, self.train_y, self.test_X, self.test_y, test_sample_index = dataset_split(full_X, self.full_y, train_proportion)
            self.test_y_sample_id = self.full_y_sample_id[test_sample_index]
        if _is_df(self.train_data):
            if _is_df(self.test_data):
                self.train_X, self.features = _reshape(self.train_X, method)
                self.test_X, self.features = _reshape(self.test_X, method)
                self.test_y_sample_id = self.test_y.index
                uniq_label = list(set(self.train_y))
                self.train_y = np.array([uniq_label.index(i) for i in self.train_y])
                self.test_y = np.array([uniq_label.index(i) for i in self.test_y])
                for i, v in enumerate(uniq_label):
                    self.num2label[i] = v
                    self.label2num[v] = i
    
    def buildmodel(self, method='default', epochs=5):
        self.model = _buildmodel(self.train_X, self.train_y, self.test_X, self.test_y, method, self.train_X.shape[-1], len(self.num2label), epochs)
    
    def explain(self, target, method='IntegratedGradients'):
        if type(self.model) is not None:
            self.attributions = _explain(self.model, self.test_X, method, target=self.label2num[target])
        attr = self.attributions.numpy()
        n_sample, n_width = attr.shape[0], attr.shape[-1]
        attr = attr.reshape(n_sample, n_width**2)
        imp = pd.DataFrame(data=attr, index=self.test_y_sample_id, columns=self.features)
        self.importances = imp
        cols = imp.apply(np.mean).abs().sort_values(ascending=False).head(20).index
        imp[cols].apply(np.mean).plot.bar().get_figure().savefig('TOP20_KeyFeatures.png', dpi=300)

    def save(self):
        self.importances.to_csv('FeatureImportance.tsv', sep='\t')
        attr_ig = np.transpose(self.attributions.squeeze().cpu().detach().numpy(), (1,2,0))
        viz.visualize_image_attr(attr_ig, sign="all", cmap="viridis", show_colorbar=True, title="", alpha_overlay=1)[0].savefig('FeatureImportances.png', dpi=300)
    


def knn(df):
    import pandas as pd
    from sklearn.impute import KNNImputer
    imputer = KNNImputer()
    data = imputer.fit_transform(X)
    return pd.DataFrame(data, index=df.index, columns=df.columns)

def minimum(df):
    df = df.T
    mins = df.min().to_dict()
    return df.fillna(mins).T

def median(df):
    df = df.T
    medians = df.median().to_dict()
    return df.fillna(medians).T

def mean(df):
    df = df.T
    means = df.mean().to_dict()
    return df.fillna(means).T

def none(df):
    return df


import sys
import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader

global W, C

class myDataset(Dataset):
    def __init__(self, dataset, labels):
        self.dataset = torch.tensor(dataset).type(torch.float32)
        self.labels = torch.tensor(labels)
    def __len__(self):
        return len(self.labels)
    def __getitem__(self, idx):
        data = self.dataset[idx]
        label = self.labels[idx]
        return data, label

class CNN(nn.Module):
    def __init__(self):
        super(CNN,self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=16, kernel_size=2, stride=1, padding=1
            ),                               #维度变换(1,28,28) --> (16,28,28)
            nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2)      #维度变换(16,28,28) --> (16,14,14)
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(in_channels=16, out_channels=32, kernel_size=2, stride=1, padding=1
            ),                               #维度变换(16,14,14) --> (32,14,14)
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2)      #维度变换(32,14,14) --> (32,7,7)
        )
        W, C = self.get_params()
        self.fc1 = nn.Linear(int(32*W), int(8*W))
        self.fc2 = nn.Linear(int(8*W), int(C))

    def forward(self, x):
        out = self.conv1(x)                  #维度变换(Batch,1,28,28) --> (Batch,16,14,14)
        out = self.conv2(out)                #维度变换(Batch,16,14,14) --> (Batch,32,7,7)
        out = out.view(out.size(0),-1)       #维度变换(Batch,32,14,14) --> (Batch,32*14*14)||将其展平
        out = self.fc1(out)
        out = self.fc2(out)
        return out
    def get_params(self):
        global W, C
        return W, C

def default_model(width, c):
    global W, C
    W = (width/4)**2
    C = c
    model = CNN()
    print(model)
    loss_fn = nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=1e-5)
    device = "cuda" if torch.cuda.is_available() else "cpu"
    print("Using {} device".format(device))
    return model, loss_fn, optimizer, device

def train(dataloader, model, loss_fn, optimizer, device):
    size = len(dataloader.dataset)
    for batch, (X, y) in enumerate(dataloader):
        X, y = X.to(device), y.to(device)
        
        # Compute prediction error
        pred = model(X)
        loss = loss_fn(pred, y)
        
        # Backpropagation
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if batch % 100 == 0:
            loss, current = loss.item(), batch * len(X)
            print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

def test(dataloader, model, loss_fn, device):
    size = len(dataloader.dataset)
    model.eval()
    test_loss, correct = 0, 0
    with torch.no_grad():
        for X, y in dataloader:
            X, y = X.to(device), y.to(device)
            pred = model(X)
            test_loss += loss_fn(pred, y).item()
            correct += (pred.argmax(1) == y).type(torch.float).sum().item()
    test_loss /= size
    correct /= size
    print(f"Test Error: \n Accuracy: {(100*correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")

def default(X_train, y_train, X_test, y_test, width, C, epochs=5):
    train_dataloader = DataLoader(myDataset(X_train, y_train), batch_size=8, shuffle=True)
    test_dataloader = DataLoader(myDataset(X_test, y_test), batch_size=8, shuffle=True)
    model, loss_fn, optimizer, device = default_model(width, C)
    for t in range(epochs):
        print(f"Epoch {t+1}\n-------------------------------")
        train(train_dataloader, model, loss_fn, optimizer, device)
        test(test_dataloader, model, loss_fn, device)
    print("Done!")
    return model

# def custom(X_train, y_train, X_test, y_test):
#     print("[Modeling] Custom: Not available yet!")
#     model = None
#     return model 

def none(X_train=None, y_train=None, X_test=None, y_test=None, width=None, C=None, epochs=None):
    print("Exit!")
    sys.exit(0)


from math import sqrt, ceil
from warnings import warn
import numpy as np


def edge(n_feature, width=None, devide=4):
    if type(width) is int:
        if width**2 >= n_feature:
            if width % devide == 0:
                return width
            else:
                warn(f"Width of {width} seems not a good choice.", Warning)
                return width
        else:
            warn(f"Width of {width} seems a bad choice. Chenged to auto mode.", Warning)
            width = None

    if width is None:
        width = ceil(sqrt(n_feature))
        while True:
            if width % devide == 0:
                return width
            else:
                width += 1

def integrate(df, patch_len):
    for i in range(patch_len):
        df['patch_'+str(i)] = 0
    return df


def auto(df):
    df = df.fillna(0)
    n_feature = len(list(df))
    n_sample = len(df.index)
    width = edge(n_feature)
    patch = width**2 - n_feature
    df = integrate(df, patch)
    features = list(df)
    ndarr = df.to_numpy()
    ndarr = ndarr.reshape(n_sample, 1, width, width)
    return ndarr, features

# def custom(df, width):
#     n_feature = len(list(df))
#     n_sample = len(df.index)
#     width = edge(n_feature, width)
#     patch_len = width**2 - n_feature
#     df = integrate(df, patch_len)
#     features = list(df)
#     ndarr = df.to_numpy()
#     ndarr = ndarr.reshape(n_sample, 1, width, width)
#     return ndarr, features

def none(df):
    ndarr = df.to_numpy()
    features = list(df)
    return ndarr, features


# import argparse

def parse():
    pass
    # parser = argparse.ArgumentParser(description='Process some integers.')
    # parser.add_argument('integers', metavar='N', type=int, nargs='+',
    #                     help='an integer for the accumulator')
    # parser.add_argument('--sum', dest='accumulate', action='store_const',
    #                     const=sum, default=max,
    #                     help='sum the integers (default: find the max)')
    # return parser.parse_args()


import numpy as np

def log2(df):
    '''x -> log_2 (x)'''
    return df.apply(np.log2, axis=1)

def loge(df):
    '''x -> log_e (x)'''
    return df.apply(np.log, axis=1)

def log2p(df):
    '''x -> log_2 (x+1)'''
    df = df.fillna(0)
    return df.apply(lambda x: np.log2(x+1), axis=1)

def logep(df):
    '''x -> log_e (x+1)'''
    df = df.fillna(0)
    return df.apply(np.log1p, axis=1)

def zscore(df):
    '''x -> (x-mean)/std'''
    return df.apply(lambda x: (x-np.mean(x))/np.std(x), axis=1)
    
def median(df):
    '''x -> x/median'''
    return df.apply(lambda x: x/np.median(x), axis=1)
    
def mean(df):
    '''x -> x/mean'''
    return df.apply(lambda x: x/np.mean(x), axis=1)

def none(df):
    '''Do nothing.'''
    return df



import torch
# IntegratedGradients, Saliency, DeepLift, DeepLiftShap, InputXGradient, GuidedBackprop, GuidedGradCam, Deconvolution, FeaturePermutation, KernelShap, LRP

try:
    import captum.attr
except:
    import sys
    print("The module 'captum' is not found, please install first.")
    print("\tconda install captum -c pytorch")
    sys.exit(0)

def explain(model, input, method='IntegratedGradients', target=2):
    interp = getattr(captum.attr, method)(model)
    attribution = interp.attribute(torch.tensor(input).type(torch.float32), target=target)
    return attribution


import pandas as pd

def read(fname, sep='\t', index_col=0):
    return pd.read_csv(fname, sep=sep, index_col=index_col)

def write(path, df, sep='\t', index=True, header=True, index_label=''):
    df.to_csv(path, sep=sep, index=index, header=header, index_label=index_label)
    return path


import miidl
from .methods import knn, minimum, median, mean, none

def impute(df, method='none'):
    return getattr(miidl.imputation, method)(df)

call = {
    'knn': knn, 
    'minimum': minimum, 
    'median': median, 
    'mean': mean,
    'none': none
}


#!/usr/bin/env python3

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
setup(
    name='miidl',
    version='0.0.2',
    description='Python package for identification of biomarkers powered by interpretable Convolutional Neural Networks',
    url='https://github.com/chunribu/miidl/',
    author='chunribu',
    author_email='chunribu@mail.sdu.edu.cn',
    packages=find_packages(),
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "torch",
        "torchvision",
        "pandas",
        "captum",
    ],
    classifiers=[
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='biomarker interpretable deeplearning',
    entry_points={
        'console_scripts': [
            # 'miidl = miidl:run_as_command',
            ]
    }
)